# -*- coding: utf-8 -*-
from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.layout.controls import UIContent, UIControl

from euporie.render import HTMLRenderer, ImageRenderer, RichRenderer, SVGRenderer

from .text import ANSI


class Control(UIControl):
    """Displays the output of a renderer re-draws it's contents on resize."""

    def __init__(self, data, render_args={}):
        self.data = data
        self.render_args = render_args

        self.renderer_instance = self.renderer.select()
        self.rendered_lines = None

        self._format_cache = SimpleCache(maxsize=20)
        self._content_cache = SimpleCache(maxsize=20)

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix,
    ):
        if not self.rendered_lines:
            self.rendered_lines = self._format_cache.get(
                (width, max_available_height),
                lambda: self.render(width, max_available_height),
            )
        return len(self.rendered_lines)

    def create_content(self, width, height):

        self.rendered_lines = self._format_cache.get(
            (width,),
            lambda: self.render(width, height),
        )

        def get_content():
            return UIContent(
                get_line=lambda i: ANSI(self.rendered_lines[i]).__pt_formatted_text__(),
                line_count=len(self.rendered_lines),
            )

        return self._content_cache.get((width,), get_content)

    def render(self, width, height):
        result = self.renderer_instance.render(
            self.data, width=width, height=height, render_args=self.render_args
        )
        if isinstance(result, str):
            result = result.rstrip().split("\n")
        return result


class RichControl(Control):
    renderer = RichRenderer


class HTMLControl(Control):
    renderer = HTMLRenderer


class ImageControl(Control):
    renderer = ImageRenderer

    def create_content(self, width, height):
        cell_obscured = self.render_args["cell"].obscured()
        self.rendered_lines = self._format_cache.get(
            (cell_obscured, width),
            lambda: self.render(width, height),
        )

        def get_content():
            return UIContent(
                get_line=lambda i: ANSI(self.rendered_lines[i]).__pt_formatted_text__(),
                line_count=len(self.rendered_lines),
            )

        return self._content_cache.get((cell_obscured, width), get_content)


class SVGControl(ImageControl):
    renderer = SVGRenderer



# -*- coding: utf-8 -*-
import logging
import sys
from functools import partial
from pathlib import Path

from prompt_toolkit.application import Application
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters import Condition, Filter
from prompt_toolkit.formatted_text import fragment_list_to_text
from prompt_toolkit.key_binding.bindings.focus import focus_next, focus_previous
from prompt_toolkit.layout import (
    ConditionalContainer,
    Float,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import LayoutDimension as D
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.layout.menus import CompletionsMenu
from prompt_toolkit.styles import BaseStyle, Style, default_ui_style, merge_styles
from prompt_toolkit.styles.pygments import style_from_pygments_cls
from prompt_toolkit.widgets import (
    Button,
    Dialog,
    Label,
    MenuContainer,
    MenuItem,
    TextArea,
)
from pygments.styles import get_all_styles, get_style_by_name

from euporie import logo
from euporie.config import config
from euporie.keys import KeyBindingsInfo
from euporie.menu import SmartMenuItem
from euporie.notebook import Notebook
from euporie.term import TermAppMixin
from euporie.text import FormatTextProcessor

log = logging.getLogger(__name__)


class App(Application, TermAppMixin):
    def __init__(self, *args, **kwargs):

        self.open_paths = []
        self.files = []
        self.last_selected_index: int = 0

        self.include_default_pygments_style = False

        self.ttimeoutlen = 0.1

        self.file_open = Condition(lambda: bool(self.files))

        self.key_binding_details = []

        super().__init__(
            layout=self.layout(),
            enable_page_navigation_bindings=True,
            mouse_support=True,
            key_bindings=self.load_key_bindings(),
            full_screen=True,
            style=None,
            editing_mode=self.get_edit_mode,
            *args,
            **kwargs,
        )

    def set_edit_mode(self, mode):
        config.editing_mode = mode
        self.editing_mode = self.get_edit_mode

    @property
    def get_edit_mode(self):
        return {"emacs": EditingMode.EMACS, "vi": EditingMode.VI}.get(
            config.editing_mode, EditingMode.EMACS
        )

    def _create_merged_style(
        self, include_default_pygments_style: Filter = None
    ) -> BaseStyle:

        # Calculate colors based on terminal background colour if we can
        if self.bg_color:
            from prompt_toolkit.styles import (
                DEFAULT_ATTRS,
                AdjustBrightnessStyleTransformation,
            )

            tr = AdjustBrightnessStyleTransformation(
                min_brightness=0.05, max_brightness=0.92
            )
            dim_bg = "#{}".format(
                tr.transform_attrs(
                    DEFAULT_ATTRS._replace(color=self.bg_color.lstrip("#"))
                ).color
            )
        else:
            dim_bg = "default"

        style_dict = {
            # The default style is merged at this point so full styles can be
            # overridden. For example, this allows us to switch off the underline
            #  status of cursor-line.
            **dict(default_ui_style().style_rules),
            "logo": "fg:#ff0000",
            "background": "fg:#888888",
            "menu-bar": "fg:#ffffff bg:#222222",
            "menu-bar.item": "bg:#444444",
            "menu": "fg:#ffffff bg:#222222",
            "cell-border": "fg:#4e4e4e",
            "cell-border-selected": "fg:#00afff",
            "cell-border-edit": "fg:#00ff00",
            "cell-input": "fg:default",
            "line-number": f"fg:#888888 bg:{dim_bg}",
            "line-number.current": "bold",
            "cursor-line": f"bg:{dim_bg}",
            "cell-output": "fg:default",
            "cell-input-prompt": "fg:darkblue",
            "cell-output-prompt": "fg:darkred",
            "scrollbar.background": "bg:#444444",
            "scrollbar.button": "bg:#aaaaaa",
            "scrollbar.arrow": "fg: #aaaaaa bg:#444444",
            "dialog shadow": "bg:#888888",
            "dialog.body": "bg:#b0b0b0 #000000",
            "hr": "fg:#666666",
        }

        # Using a dynamic style has serious performance issues, so instead we update
        # the style on the renderer directly when it changes in `self.update_style`
        return merge_styles(
            [
                Style.from_dict(style_dict),
                style_from_pygments_cls(get_style_by_name(config.pygments_style)),
            ]
        )

    def update_style(self, pygments_style):
        config.pygments_style = pygments_style
        self.renderer.style = self._create_merged_style()

    def layout(self):

        kernel_status_repr = {
            "starting": "◍",
            "idle": "○",
            "busy": "●",
            "error": "☹",
        }

        def get_statusbar_text():
            kernel_status = kernel_status_repr.get(self.file.kernel_status, "◌")
            selected_cell = self.file.page.selected_index + 1
            dirt = "*" if self.file.dirty else ""
            return [
                ("class:menu-bar.item", f" Cell {selected_cell} "),
                ("", " "),
                ("class:menu-bar.item bold", f" {dirt}{self.file.path.name} "),
                ("", " "),
                ("class:menu-bar.item", f" {self.file.kernel_name} {kernel_status} "),
            ]

        status_bar_text = FormattedTextControl(
            get_statusbar_text, focusable=True, show_cursor=False
        )
        self.status_bar = ConditionalContainer(
            Window(
                content=status_bar_text,
                height=D.exact(1),
                style="class:menu-bar",
                dont_extend_width=True,
                align=WindowAlign.RIGHT,
            ),
            filter=self.file_open,
        )

        self.body_container = MenuContainer(
            body=HSplit([VSplit(self.files)]),
            menu_items=[
                MenuItem(
                    " File ",
                    children=[
                        MenuItem(
                            "New", handler=lambda: self.ask_open_file(validate=False)
                        ),
                        MenuItem("Open", handler=self.ask_open_file),
                        MenuItem("-", disabled=True),
                        SmartMenuItem(
                            "Save",
                            handler=lambda: self.file.save(),
                            disabler=~self.file_open,
                        ),
                        SmartMenuItem(
                            "Close", handler=self.close_file, disabler=~self.file_open
                        ),
                        MenuItem("-", disabled=True),
                        MenuItem("Exit", handler=self.try_to_exit),
                    ],
                ),
                MenuItem(
                    " Edit ",
                    children=[
                        MenuItem("Cut Cell", handler=lambda: self.file.cut()),
                        MenuItem("Copy Cell", handler=lambda: self.file.copy()),
                        MenuItem("Paste Cell", handler=lambda: self.file.paste()),
                    ],
                ),
                MenuItem(
                    " Kernel ",
                    children=[
                        MenuItem(
                            "Restart Kernel", handler=lambda: self.file.restart_kernel()
                        ),
                        MenuItem(
                            "Change Kernel...",
                            handler=lambda: self.file.change_kernel(),
                        ),
                    ],
                ),
                MenuItem(
                    " Settings ",
                    children=[
                        MenuItem(
                            "Syntax Theme",
                            children=[
                                SmartMenuItem(
                                    style,
                                    handler=partial(self.update_style, style),
                                    toggler=Condition(
                                        partial(
                                            lambda x: config.pygments_style == x, style
                                        )
                                    ),
                                )
                                for style in sorted(get_all_styles())
                            ],
                        ),
                        MenuItem(
                            "Editing Key Bindings",
                            children=[
                                SmartMenuItem(
                                    "Emacs",
                                    handler=lambda: self.set_edit_mode("emacs"),
                                    toggler=Condition(
                                        lambda: config.editing_mode == "emacs"
                                    ),
                                ),
                                SmartMenuItem(
                                    "Vi",
                                    handler=lambda: self.set_edit_mode("vi"),
                                    toggler=Condition(
                                        lambda: config.editing_mode == "vi"
                                    ),
                                ),
                            ],
                        ),
                        MenuItem(
                            "Switch Background Pattern",
                            handler=lambda: config.toggle("background"),
                        ),
                        SmartMenuItem(
                            "Run cell after external edit",
                            toggler=Condition(
                                lambda: config.execute_after_external_edit
                            ),
                            handler=lambda: config.toggle(
                                "execute_after_external_edit"
                            ),
                        ),
                    ],
                ),
                MenuItem(
                    " Help ",
                    children=[
                        MenuItem("Keyboard Shortcuts", handler=self.help_keys),
                        MenuItem("About", handler=self.help_about),
                    ],
                ),
            ],
        )

        self.logo = FormattedTextControl(
            [("", f" {logo} ")],
            focusable=True,
            show_cursor=False,
            style="class:menu-bar.item,logo",
        )

        # Add logo and status bar into menubar
        self.body_container.container.content.children = [
            VSplit(
                [
                    Window(self.logo, height=D.exact(1), dont_extend_width=True),
                    self.body_container.window,
                    self.status_bar,
                ]
            ),
            self.body_container.container.content.children[1],
        ]

        self.root_container = FloatContainer(
            content=self.body_container,
            floats=[
                Float(
                    xcursor=True,
                    ycursor=True,
                    content=CompletionsMenu(max_height=16, scroll_offset=1),
                )
            ],
        )

        # Focus status_bar_text so notebook is selected on tab
        return Layout(self.root_container, focused_element=status_bar_text)

    def register_key_binding(self, kb, scope, desc, *keys, **kwargs):
        self.key_binding_details.append((scope, desc, keys))
        return kb.add(*keys, **kwargs)

    def load_key_bindings(self):
        kb = KeyBindingsInfo()

        in_edit_mode = Condition(
            lambda: self.file and self.file.cell and self.file.cell.is_editing()
        )

        @kb.add("c-n", group="Application", desc="Create a new notebook file")
        def new(event):
            self.ask_open_file(validate=False)

        @kb.add("c-o", group="Application", desc="Open file")
        def open(event):
            self.ask_open_file()

        @kb.add("c-w", group="Application", desc="Close the current file")
        def close(event):
            self.close_file(self.file)

        @kb.add("c-q", group="Application", desc="Quit euporie")
        def exit(event):
            self.try_to_exit()

        kb.add(
            "tab", filter=~in_edit_mode, group="Navigation", desc="Focus next element"
        )(focus_next)
        kb.add(
            "s-tab",
            filter=~in_edit_mode,
            group="Navigation",
            desc="Focus previous element",
        )(focus_previous)

        return kb

    @property
    def selected_index(self):
        # Detect if focused file has changed
        # Find index of selected child
        for i, file in enumerate(self.files):
            if self.layout.has_focus(file):
                break
        else:
            i = self.last_selected_index
        # This will perform change the position when the new child is selected
        self.last_selected_index = i
        return self.last_selected_index

    @selected_index.setter
    def selected_index(self, new_index):
        # Only update the selected file if it was not selected before
        if new_index != self.last_selected_index:
            # Ensure selected index is a valid child
            new_index = min(max(new_index, 0), len(self.files) - 1)
            self.select_notebook(new_index)

            # Focus new child if not already focused
            child = self.get_child(new_index)
            if not self.layout.has_focus(child):
                self.layout.focus(child)

            # Track which child was selected
            self.last_selected_index = new_index

    def try_to_exit(self):
        """
        Check for unsaved files before closing.

        Creates a chain of close file commands, where the callback for each triggers
        the closure of the next. The closing process can be cancelled anywhere along
        the chain.
        """

        if self.files:

            def final_cb():
                """Really exit after the last file in the chain is closed."""
                self.cleanup_closed_file(self.files[0])
                self.exit()

            def create_cb(close_file, cleanup_file, cb):
                """Generate a file close chaining callbacks.

                Cleans up after the previously closed file, and requests to close the
                next file in the chain.
                """

                def inner():
                    self.cleanup_closed_file(cleanup_file)
                    close_file.close(cb=cb)

                return inner

            cb = final_cb
            for close_file, cleanup_file in zip(self.files, self.files[1:]):
                cb = create_cb(close_file, cleanup_file, cb)
            self.files[-1].close(cb)
        else:
            self.exit()

    def ask_open_file(self, default="", validate=True, error=None):
        def open_cb():
            path = filepath.text
            if not validate or Path(path).expanduser().exists():
                self.open_file(filepath.text)
            else:
                self.ask_open_file(
                    default=filepath.text, validate=validate, error="File not found"
                )

        filepath = TextArea(text=default, multiline=False)
        self.dialog(
            title="Select file",
            body=HSplit(
                [
                    Label("Enter file name:"),
                    filepath,
                ]
                + ([Label(error, style="red")] if error else [])
            ),
            buttons={
                "OK": open_cb,
                "Cancel": None,
            },
            to_focus=filepath,
        )

    def open_file(self, path):
        path = Path(path).expanduser()
        log.info(f"Opening file {path}")
        open_paths = [x.path for x in self.files]
        if path in open_paths:
            log.info(f"File {path} already open, activating")
            self.active_notebook = open_paths.index(path)
        else:
            self.open_paths.append(path)
            self.files.append(Notebook(path))
            self.active_notebook_index = len(self.files)

        self.body_container.body.children[0] = VSplit(self.files)

    def close_file(self, file=None):
        if file is None:
            file = self.file
        if file:
            self.file.close(lambda: self.cleanup_closed_file(file))

    def cleanup_closed_file(self, file):
        # Remove file
        self.files.remove(file)
        # Update body container to reflect new file list
        self.body_container.body.children[0] = VSplit(self.files)
        # Focus another file if one exists
        if self.files:
            self.file.page.focus()
        # If a file is not open, the status bar is not shown, so focus the logo, so
        # pressing tab focuses the menu
        else:
            self.layout.focus(self.logo)

    @property
    def file(self):
        if self.files:
            index = min(self.selected_index, len(self.files) - 1)
            return self.files[index]

    def dialog(self, title, body, buttons, to_focus=None):
        def make_handler(cb):
            def inner():
                self.root_container.floats.remove(dialog)
                self.layout.focus(focused)
                if cb:
                    cb()

            return inner

        focused = self.layout.current_control

        button_widgets = [
            Button(text, make_handler(cb), left_symbol="[", right_symbol="]")
            for text, cb in buttons.items()
        ]

        dialog = Float(
            Dialog(
                title=title,
                body=body,
                buttons=button_widgets,
                modal=True,
                with_background=True,
            )
        )
        self.root_container.floats.append(
            dialog,
        )
        if to_focus is None:
            to_focus = button_widgets[0]
        self.layout.focus(to_focus)

    def help_keys(self):

        key_details = {
            group: {
                " / ".join(
                    [
                        " ".join(
                            (
                                part.replace("c-", "ctrl-").replace("s-", "shift-")
                                for part in key
                            )
                        )
                        for key in keys
                    ]
                ): desc
                for desc, keys in info.items()
            }
            for group, info in KeyBindingsInfo.details.items()
        }
        max_key_len = (
            max([len(key) for group in key_details.values() for key in group]) + 1
        )

        fragment_list = []
        for group, item in key_details.items():
            fragment_list.append(("", " " * (max(0, max_key_len - len(group)))))
            fragment_list.append(("bold underline", f"{group}\n"))
            for key, desc in item.items():
                fragment_list.append(("bold", key.rjust(max_key_len)))
                fragment_list.append(("", f"  {desc}\n"))
            fragment_list.append(("", "\n"))
        fragment_list.pop()

        plain_text = fragment_list_to_text(fragment_list)
        body = TextArea(
            text=plain_text.strip(),
            multiline=True,
            focusable=True,
            wrap_lines=False,
            input_processors=[FormatTextProcessor()],
            width=max([len(line) for line in plain_text.split("\n")]) + 2,
            scrollbar=True,
        )
        body.control.formatted_text = fragment_list

        self.dialog(
            title="Keyboard Shortcuts",
            body=body,
            buttons={"OK": None},
        )

    def help_about(self):
        self.dialog(
            title="About",
            body=Window(
                FormattedTextControl(
                    [
                        ("class:logo", logo),
                        ("", " "),
                        ("bold", "euporie\n\n"),
                        ("", "A command line editor for Jupyter notebooks\n"),
                        ("class:hr", "─" * 43 + "\n\n"),
                        ("", "Created by Josiah Outram Halstead\n"),
                        ("", "© 2021"),
                    ]
                ),
                dont_extend_height=True,
            ),
            buttons={"OK": None},
        )

    @classmethod
    def launch(cls):
        app = cls()
        for path in sys.argv[1:]:
            app.open_file(path)
        app.run()



# -*- coding: utf-8 -*-
import array
import re
import sys
from functools import lru_cache

CONTROL_RE = re.compile(
    r"""
\x1b
(?:
 (?P<csi>  # Control Sequence Introducer
  \[
  (?P<csi_params>[0-9:;<=>?]*)
  (?P<csi_inter>[!\"#$%&'()*+,\-./]*)
  (?P<csi_final>[A-Za-z@[\]^_`a–z{|}~])
 )
 |
 (?P<osc>  # Operating System Command
  \]
  (?P<osc_string>.*(?=\x1b\\|\x07))
  (?P<osc_term>\x1b\\|\x07) # Terminated by ST or BEL
 )
 |
 (?P<apc>  # Application Program Command
  _
  (?P<apc_string>.*(?=\x1b\\))
  (?P<apc_term>\x1b\\) # Terminated by ST or BEL
 )
 |
 (?P<fe> # Other Fe escape sequences
  [NOP[\\\]X^_]
 )
)
    """,
    re.VERBOSE,
)


class QueryCodes:
    device = "\x1b[>0c"
    pixel_dimensions = "\x1b[14t"
    background_color = "\x1b]11;?\x1b\\"
    sixel = "\033[?1;1;0S"
    # For some reason, konsole prints APC codes to the screen, so we clear the current
    # line and move the curor to the start after sending a kitty graphics query code,
    # which prevents breaking the display
    kitty = "\x1b_Gi=1,s=1,v=1,a=q,t=d,f=24;AAAA\x1b\\" + "\x1b[1K" + "\x1b[1G"


class QueryResponsePatterns:
    device = re.compile(r"\x1b\[\>(\d;?)+c")
    background_color = re.compile(
        "11;rgb:(?P<r>[0-9A-Fa-f]{2,4})/(?P<g>[0-9A-Fa-f]{2,4})/(?P<b>[0-9A-Fa-f]{2,4})"
    )
    pixel_dimensions = re.compile(r"4;(?P<y>\d+);(?P<x>\d+)")
    sixel = re.compile(r"\d+")
    kitty = "\x1b[_Gi=1;OK\x1b[\\"


class TermAppMixin:
    @property
    @lru_cache
    def _have_termios_tty_fcntl(self):
        try:
            import fcntl  # noqa F401
            import termios  # noqa F401
            import tty  # noqa F401
        except ModuleNotFoundError:
            return False
        else:
            return True

    def _query_term(self, query_code):
        """
        We use a Secondary Device Attribute request as our deliminator to locate the
        query response string
        """
        if self._have_termios_tty_fcntl:
            import termios
            import tty

            stdin = sys.stdin.fileno()
            tattr = termios.tcgetattr(stdin)

            output = ""

            try:
                tty.setcbreak(stdin, termios.TCSANOW)
                self.output.write_raw(QueryCodes.device)
                self.output.flush()

                output_started = False
                query_sent = False

                while True:
                    output += sys.stdin.read(1)
                    # Send query after first character of first deliminator recieved
                    if not query_sent:
                        self.output.write_raw(query_code)
                        self.output.flush()
                        query_sent = True
                    if re.search(QueryResponsePatterns.device, output):
                        if not output_started:
                            output_started = True
                            # Discard the first deliminator
                            output = ""
                            # Send the second deliminator
                            self.output.write_raw(QueryCodes.device)
                            self.output.flush()
                            continue
                        break
            finally:
                termios.tcsetattr(stdin, termios.TCSANOW, tattr)
            # Remove deliminator
            output = re.sub(QueryResponsePatterns.device, "", output, 1)
            # Parse result
            if match := CONTROL_RE.search(output):
                return match.groupdict()

    def tiocgwnsz(self):
        output = array.array("H", [0, 0, 0, 0])
        if self._have_termios_tty_fcntl:
            import fcntl
            import termios

            fcntl.ioctl(1, termios.TIOCGWINSZ, output)
        rows, cols, xpixels, ypixels = output
        return rows, cols, xpixels, ypixels

    @property
    def bg_color(self):
        if result := self._query_term(QueryCodes.background_color):
            if oscs := result.get("osc_string"):
                if match := QueryResponsePatterns.background_color.match(oscs):
                    if colors := match.groupdict():
                        if len(colors) >= 3:
                            r, g, b = colors.get("r"), colors.get("g"), colors.get("b")
                            return f"#{r[:2]}{g[:2]}{b[:2]}"

    @property
    def term_size_px(self):
        *_, px, py = self.tiocgwnsz()
        # If unsuccessful, try requesting info with escape code method
        if px == 0:
            if result := self._query_term(QueryCodes.pixel_dimensions):
                params = result.get("csi_params", "")
                if match := QueryResponsePatterns.pixel_dimensions.match(params):
                    if values := match.groupdict():
                        px, py = int(values.get("x")), int(values.get("y"))
        return px, py

    @property
    def char_size_px(self):
        px, py = self.term_size_px
        rows, cols = self.output.get_size()
        # If we can't get the pixel size, just guess wildly
        if px == 0:
            return 10, 22
        return px // cols, py // rows

    @property
    @lru_cache
    def has_sixel_graphics(self):
        if result := self._query_term(QueryCodes.sixel):
            params = result.get("csi_params", "")
            params = QueryResponsePatterns.sixel.findall(params)
            params = [int(x) for x in params]
            if len(params) >= 3:
                Pi, Ps, Pv = params[:3]
                if Ps == 0:
                    return True
        return False

    @property
    @lru_cache
    def has_kitty_graphics(self):
        if result := self._query_term(QueryCodes.kitty):
            apc_string = result.get("apc_string", "")
            if apc_string.startswith("G"):
                if len(response := apc_string.lstrip("G").split(";")) >= 2:
                    if response[1] == "OK":
                        return True
        return False



# -*- coding: utf-8 -*-
from typing import Any

import rich.markdown
from rich.console import Console, ConsoleOptions, RenderResult
from rich.table import Table as RichTable

from euporie.config import config
from euporie.libs.commonmark_extensions.tables import ParserWithTables

rich.markdown.Parser = ParserWithTables


class Table(rich.markdown.MarkdownElement):
    """A table."""

    style_name = "markdown.table"
    new_line = False

    @classmethod
    def create(cls, markdown: rich.markdown.Markdown, table: Any) -> "Table":
        return cls(table)

    def __init__(self, contents) -> None:
        self.contents = contents.table
        self.column_properties = contents.column_properties
        self.elements = []

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        table = RichTable()

        def node_to_text(node, justify=None):
            """Create a rich markdown object, then add the parsed node to it."""
            node_md = rich.markdown.Markdown(
                "",
                code_theme=config.pygments_style,
                justify=justify,
            )
            node_md.parsed = node
            return node_md

        for i, col in enumerate(self.contents[0][0]):
            justify = self.column_properties.get(i, {}).get("align")
            table.add_column(node_to_text(col, justify=justify), justify=justify)

        for row in self.contents[1]:
            table.add_row(*[node_to_text(cell, justify=justify) for cell in row])

        yield table


rich.markdown.Markdown.elements["table"] = Table



# -*- coding: utf-8 -*-
from pathlib import PurePath

from prompt_toolkit.layout import Container, Window, to_container
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.widgets import Label
from rich.markdown import Markdown

from euporie import mdtable  # noqa E401
from euporie.config import config
from euporie.control import HTMLControl, ImageControl, RichControl, SVGControl
from euporie.text import ANSI

BLING_SCORES = {
    "image/*": 0,
    "text/html": 1,
    "text/x-markdown": 2,
    "text/x-python-traceback": 3,
    "text/stderr": 4,
    "text/*": 5,
    "*": 6,
}


def bling(item):
    mime, _ = item
    for bling_path, score in BLING_SCORES.items():
        if PurePath(mime).match(bling_path):
            return score


class Output:
    def __init__(self, index, json, parent=None):
        self.index = index
        self.json = json
        self.parent = parent
        self.style = ""
        self.content = None

        # Sort data first so there is more bling first
        for mime, datum in sorted(self.data.items(), key=bling):

            mime_path = PurePath(mime)

            if mime_path.match("image/svg+xml"):
                control = SVGControl(
                    datum,
                    render_args=dict(cell=self.parent, output_index=index),
                )
                if control.renderer:
                    self.content = Window(control, height=control.rendered_lines)
                    break

            if mime_path.match("image/*"):
                control = ImageControl(
                    datum,
                    render_args=dict(cell=self.parent, output_index=index),
                )
                if control.renderer:
                    self.content = Window(control, height=control.rendered_lines)
                    break

            if mime_path.match("text/html"):
                control = HTMLControl(datum)
                if control.renderer:
                    self.content = Window(control, height=control.rendered_lines)
                    break
                else:
                    continue  # Use plain text rendering instead

            if mime_path.match("text/x-markdown"):
                control = RichControl(
                    Markdown(
                        datum,
                        code_theme=config.pygments_style,
                        inline_code_theme=config.pygments_style,
                    )
                )
                self.content = Window(control, height=control.rendered_lines)
                break

            if mime_path.match("text/x-python-traceback"):
                control = FormattedTextControl(ANSI(datum))
                self.content = Window(control)
                break

            if mime_path.match("text/stderr"):
                control = FormattedTextControl(ANSI(datum.rstrip()))
                self.content = Window(control, wrap_lines=True, style="fg:red")
                break

            if mime_path.match("text/*"):
                # Use formatted text so ansi colour codes are displayed as colours
                control = FormattedTextControl(ANSI(datum.rstrip()))
                self.content = Window(control, wrap_lines=True)
                break

        else:
            datum = sorted(self.data.items(), key=bling)[-1][1]
            self.content = Label(str(datum).rstrip())

    @property
    def data(self):
        output_type = self.json.get("output_type", "unknown")
        if output_type == "stream":
            return {f'text/{self.json.get("name")}': self.json.get("text")}
        elif output_type == "error":
            return {"text/x-python-traceback": "\n".join(self.json.get("traceback"))}
        else:
            return self.json.get("data", {})

    @property
    def metadata(self):
        return self.json.get("metadata", {})

    def get_key_bindings(self, *args, **kwargs):
        return to_container(self.content).get_key_bindings(*args, **kwargs)

    def get_children(self, *args, **kwargs):
        return to_container(self.content).get_children(*args, **kwargs)

    def preferred_width(self, *args, **kwargs):
        return to_container(self.content).preferred_width(*args, **kwargs)

    def preferred_height(self, *args, **kwargs):
        return to_container(self.content).preferred_height(*args, **kwargs)

    def write_to_screen(self, *args, **kwargs):
        return to_container(self.content).write_to_screen(*args, **kwargs)

    def reset(self, *args, **kwargs):
        return to_container(self.content).reset(*args, **kwargs)

    def __pt_container__(self) -> Container:
        return to_container(self.content)



# -*- coding: utf-8 -*-
import json
import logging
import os
import sys
from functools import lru_cache
from pathlib import Path

import jsonschema

from euporie import _app_name

log = logging.getLogger(__name__)

CONFIG_SCHEMA = {
    "title": "Euporie Configuration",
    "description": "A configuration for euporie",
    "type": "object",
    "properties": {
        "max_notebook_width": {
            "description": "The maximum width of the notebook",
            "type": "integer",
            "minimum": 1,
            "default": 120,
        },
        "background": {
            "description": "The background pattern to use",
            "type": "integer",
            "minimum": 0,
            "maximum": 4,
            "default": 1,
        },
        "background_character": {
            "description": "The character to use to draw the background",
            "type": "string",
            "maxLength": 1,
            "default": "·",
        },
        "pygments_style": {
            "description": "The name of the pygments style for syntax highlighting",
            "type": "string",
            "default": "default",
        },
        "editing_mode": {
            "description": "The key-binding style to use for text editing",
            "type": "string",
            "pattern": "(emacs|vi)",
            "default": "emacs",
        },
        "show_line_numbers": {
            "description": "Whether line numbers are shown by default",
            "type": "boolean",
            "default": True,
        },
        "execute_after_external_edit": {
            "description": "Whether to execute a cell immediately after editing in $EDITOR",
            "type": "boolean",
            "default": False,
        },
    },
}


class Config:
    win = sys.platform.startswith("win") or (sys.platform == "cli" and os.name == "nt")

    def __init__(self):
        super().__setattr__("valid", True)
        self.load_defaults()
        self.load_config_file()

    @property
    @lru_cache
    def config_file_path(self):
        if self.win:
            path = (
                Path(os.getenv("APPDATA", "~/appdata")).expanduser()
                / "roaming"
                / _app_name
            )
        elif sys.platform == "darwin":
            path = Path("~/Library/Application Support/").expanduser() / _app_name
        else:
            path = (
                Path(os.getenv("XDG_CONFIG_HOME", "~/.config")).expanduser() / _app_name
            )
        # Create config folder if it doesn't exist
        path.mkdir(exist_ok=True, parents=True)
        return path / "config.json"

    @property
    def json_data(self):
        if self.valid and self.config_file_path.exists():
            with open(self.config_file_path, "r") as f:
                try:
                    json_data = json.load(f)
                except json.decoder.JSONDecodeError:
                    log.error(
                        "Could not parse the configuration file: "
                        f"{self.config_file_path}\n"
                        "Is it valid json?"
                    )
                    super().__setattr__("valid", False)
                else:
                    try:
                        jsonschema.validate(instance=json_data, schema=CONFIG_SCHEMA)
                    except jsonschema.ValidationError as error:
                        log.error(
                            f"Error in config file: {self.config_file_path}\n"
                            f"{error}"
                        )
                        super().__setattr__("valid", False)
                    else:
                        return json_data
            log.warning("The configuration file was not loaded")
        return {}

    def set(self, attr, value):
        if hasattr(self, attr):
            setattr(self, attr, value)

    def toggle(self, attr):
        if hasattr(self, attr):
            current = getattr(self, attr)
            schema = CONFIG_SCHEMA["properties"][attr]
            if schema["type"] == "boolean":
                setattr(self, attr, not current)
            elif schema["type"] == "integer":
                setattr(
                    self,
                    attr,
                    schema["minimum"]
                    + (current - schema["minimum"] + 1) % (schema["maximum"] + 1),
                )

    def load_defaults(self):
        for key, schema in CONFIG_SCHEMA["properties"].items():
            super().__setattr__(key, schema["default"])

    def load_config_file(self):
        for key, value in self.json_data.items():
            super().__setattr__(key, value)

    def __setattr__(self, name, value):
        if self.valid:
            json_data = self.json_data
            json_data[name] = value
            with open(self.config_file_path, "w") as f:
                json.dump(json_data, f, indent=2)
        return super().__setattr__(name, value)


config = Config()



# -*- coding: utf-8 -*-
import logging

from prompt_toolkit.patch_stdout import StdoutProxy
from rich.console import Console
from rich.logging import RichHandler

log_stdout = StdoutProxy(raw=True)

handlers = [
    # logging.StreamHandler(log_stdout),
    RichHandler(
        console=Console(file=log_stdout, force_terminal=True),
        markup=True,
        rich_tracebacks=True,
    )
]

logging.basicConfig(
    level=logging.WARNING,
    handlers=handlers,
    format="%(message)s",
)



# -*- coding: utf-8 -*-
from typing import Callable, Optional, Union

from prompt_toolkit.filters import FilterOrBool
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.keys import Keys


class KeyBindingsInfo(KeyBindings):
    details = {}

    def add(
        self,
        *keys: Union[Keys, str],
        filter: FilterOrBool = True,
        eager: FilterOrBool = False,
        is_global: FilterOrBool = False,
        save_before: Callable[["KeyPressEvent"], bool] = (lambda e: True),
        record_in_macro: FilterOrBool = True,
        key_str: Optional[str] = None,
        group: Optional[str] = "None",
        desc: Optional[str] = None,
    ):
        if desc is not None:
            self.details.setdefault(group, {}).setdefault(desc, {})[
                key_str or keys
            ] = None
        return super().add(
            *keys,
            filter=filter,
            eager=eager,
            is_global=is_global,
            save_before=save_before,
            record_in_macro=record_in_macro,
        )



# -*- coding: utf-8 -*-
import base64
import io
import os
import subprocess
import tempfile
from importlib import import_module
from math import ceil
from shutil import which

import rich
from PIL import Image
from prompt_toolkit.application import get_app

from euporie.box import Border


class DataRenderer:
    def __init__(self, **kwargs):

        self.width = 1
        self.height = 1
        for kwarg, value in kwargs.items():
            setattr(self, kwarg, value)
        self.load()

    def load(self):
        pass

    # async def _render(self, mime, data, **kwargs):
    def render(self, data, width, height, render_args=None):

        self.width = width
        self.height = height
        if render_args is None:
            render_args = {}
        self.render_args = render_args

        self.load()
        ansi_data = self.process(data)
        # data = await self.process(data)
        return ansi_data

    @classmethod
    def select(cls, *args, **kwargs):
        sub_renderers = cls.__subclasses__()
        if sub_renderers:
            for Renderer in sub_renderers:
                renderer = Renderer(*args, **kwargs)
                if not hasattr(renderer, "validate"):
                    # It's just a grouping class
                    continue
                if renderer.validate():
                    return renderer
        else:
            renderer = cls()
            if renderer.validate():
                return renderer


class SubprocessRenderMixin:
    use_tempfile = False
    to_string = True

    def validate(self):
        return which(self.cmd[0])

    def process(self, data):
        # async def process(self, data):

        cmd = list(map(str, self.cmd))

        if isinstance(data, str):
            data = data.encode()

        if self.use_tempfile:
            # If the command cannot read from stdin, create a temporary file to pass to
            # the command
            tfile = tempfile.NamedTemporaryFile(delete=False)
            tfile.write(data)
            tfile.close()
            cmd.append(tfile.name)

        # TODO render asynchronously
        # proc = await asyncio.create_subprocess_shell(
        # " ".join(cmd),
        # stdout=asyncio.subprocess.PIPE,
        # stdin=asyncio.subprocess.PIPE,
        # stderr=asyncio.subprocess.DEVNULL,
        # )
        # stdout, stderr = await proc.communicate(data)

        cmd = list(cmd)
        output = subprocess.check_output(cmd, input=data)

        # TODO Log any stderr
        # print(stderr)

        # Clean up any temporary file
        if self.use_tempfile:
            tfile.close()
            os.unlink(tfile.name)

        if self.to_string:
            output = output.decode()

        return output


class PythonRenderMixin:
    def validate(self):
        try:
            import_module(self.module)  # noqa
        except ModuleNotFoundError:
            return False
        else:
            return True


class RichRendererMixin(DataRenderer):
    def load(self):
        self.console = rich.get_console()

    def process(self, data):
        buffer = self.console.render(
            data,
            self.console.options.update(max_width=self.width),
        )
        self.rendered_lines = self.console._render_buffer(buffer)
        return self.rendered_lines


class RichRenderer(RichRendererMixin):
    def validate(self):
        return True


class HTMLRenderer(DataRenderer):
    pass


class html_w3m(HTMLRenderer, SubprocessRenderMixin):
    def load(self):
        self.cmd = ["w3m", "-T", "text/html", "-cols", f"{self.width}"]


class html_elinks(HTMLRenderer, SubprocessRenderMixin):
    def load(self):
        self.cmd = [
            "elinks",
            "-dump",
            "-dump-width",
            f"{self.width}",
            "-no-numbering",
            "-force-html",
            "-no-references",
        ]


class html_lynx(HTMLRenderer, SubprocessRenderMixin):
    def load(self):
        self.cmd = ["lynx", "-dump", "-stdin", f"-width={self.width}"]


class html_links(HTMLRenderer, SubprocessRenderMixin):
    use_tempfile = True

    def load(self):
        self.cmd = ["links", "-width", self.width, "-dump"]


class html_mtable_py(RichRendererMixin, HTMLRenderer, PythonRenderMixin):
    """Converts HTML tables to markdown, then renders markdown with rich."""

    module = "mtable"

    def process(self, data):
        from mtable import MarkupTable

        return super().process(
            rich.markdown.Markdown(
                "\n\n".join([table.to_md() for table in MarkupTable.from_html(data)])
            )
        )


class html_fallback_py(HTMLRenderer):
    """
    This uses stdlib to strip html tags.

    This is the last resort for rendering HTML.
    """

    stripper = None

    def validate(self):
        return True

    def load(self):
        from html.parser import HTMLParser

        if self.stripper is None:

            class HTMLStripper(HTMLParser):
                """Very basic HTML parser which strips style and script tags."""

                def __init__(self):
                    super().__init__()
                    self.reset()
                    self.strict = False
                    self.convert_charrefs = True
                    self.text = io.StringIO()
                    self.skip = False

                def handle_starttag(self, tag, attrs):
                    if tag in ("script", "style"):
                        self.skip = True

                def handle_endtag(self, tag):
                    if tag in ("script", "style"):
                        self.skip = False

                def handle_data(self, d):
                    if not self.skip:
                        self.text.write(d)

                def get_data(self):
                    return self.text.getvalue()

            self.stripper = HTMLStripper()

    def process(self, data):
        import re

        self.stripper.feed(data)
        data = self.stripper.get_data()
        data = "\n".join([x.strip() for x in data.strip().split("\n")])
        data = re.sub("\n\n\n+", "\n\n", data)
        return data


class ImageRenderer(DataRenderer):
    def __init__(self, *args, image=None, **kwargs):
        self.image = image
        self.px = self.py = 0
        super().__init__(*args, **kwargs)

    def render(self, data, width=None, height=None, render_args=None):

        img_bytes = io.BytesIO(base64.b64decode(data))
        try:
            self.image = Image.open(img_bytes)
        except IOError:
            pass
        else:
            orig_px, orig_py = self.image.size

            # Scale image down if it is larger than available width
            app = get_app()
            if hasattr(app, "app"):
                char_px = app.app.char_px
                char_py = app.app.char_py
            else:
                char_px, char_py = get_app().char_size_px

            pixels_per_col = orig_px / char_px
            # Only down-scale images
            scaling_factor = min(1, width / pixels_per_col)
            # Pixel & character values need to be integers
            self.px = ceil(orig_px * scaling_factor)
            self.py = ceil(orig_py * scaling_factor)

            width = ceil(self.px / char_px)
            height = ceil(self.py / char_py)

        return super().render(data, width=width, height=height, render_args=render_args)


class SixelMixerRenderer(ImageRenderer):
    def validate(self):
        return get_app().has_sixel_graphics

    def process(self, data):
        """Add in cursor movement commands to correctly place sixels"""
        output = ""
        # Add text representation
        if self.render_args.get("cell").obscured():
            output += AnsiImageRenderer.select(image=self.image).render(
                data,
                width=self.width,
                height=self.height,
                render_args=self.render_args,
            )
        else:
            output += "\n".join([" " * self.width] * self.height)

        # data = super().process(data)
        # Save cursor position
        output += "\x1b[s"
        # Move cursor back to top left of image's space
        output += f"\x1b[{self.height-1}A\x1b[{self.width}D"
        # Add sixels
        # output += super().process(data)
        sixel = SixelRenderer.select(image=self.image).render(
            data,
            width=self.width,
            height=self.height,
            render_args=self.render_args,
        )
        output += sixel
        # Restore cursor position
        output += "\x1b[u"
        return output


class SixelRenderer(DataRenderer):
    def validate(self):
        return get_app().has_sixel_graphics and super().validate()


class img_sixel_imagemagik(SixelRenderer, ImageRenderer, SubprocessRenderMixin):
    def load(self):
        if not hasattr(self, "px"):
            self.px = self.py = 0
            self.bg_color = "#FFFFFF"
        self.cmd = [
            "convert",
            "-",
            "-geometry",
            f"{self.px}x{self.py}",
            "-background",
            get_app().bg_color,
            "-flatten",
            "sixel:-",
        ]

    def process(self, data):
        data = base64.b64decode(data)
        data = super().process(data)
        return data


class img_sixel_timg_py(SixelRenderer, ImageRenderer, PythonRenderMixin):
    module = "timg"

    def process(self, data):
        import timg

        self.image.thumbnail((self.px, self.px))
        # Set transparent colour to terminal background
        if self.image.mode in ("RGBA", "LA") or (
            self.image.mode == "P" and "transparency" in self.image.info
        ):
            alpha = self.image.convert("RGBA").getchannel("A")
            bg = Image.new("RGBA", self.image.size, get_app().bg_color)
            bg.paste(self.image, mask=alpha)
            self.image = bg
        self.image = self.image.convert("P", palette=Image.ADAPTIVE, colors=16).convert(
            "RGB", palette=Image.ADAPTIVE, colors=16
        )
        data = timg.SixelMethod(self.image).to_string()
        return data


class img_sixel_teimpy(SixelRenderer, ImageRenderer, PythonRenderMixin):
    module = "teimpy"

    def process(self, data):
        import numpy as np
        import teimpy

        # data = super().process(data)
        self.image.thumbnail((self.px, self.px))
        # Set transparent colour to terminal background
        if self.image.mode in ("RGBA", "LA") or (
            self.image.mode == "P" and "transparency" in self.image.info
        ):
            alpha = self.image.convert("RGBA").getchannel("A")
            bg = Image.new("RGBA", self.image.size, get_app().bg_color)
            bg.paste(self.image, mask=alpha)
            self.image = bg.convert("RGB")
        data = teimpy.get_drawer(teimpy.Mode.SIXEL).draw(np.asarray(self.image))
        return data


class img_kitty(ImageRenderer):
    def validate(self):
        return get_app().has_kitty_graphics

    def load(self):
        pass

    def process(self, data):

        output = ""

        # Create image id for kitty using Cantor pairing of cell id and output index
        a = self.render_args.get("cell").index
        b = self.render_args.get("output_index")
        image_id = int(0.5 * (a + b) * (a + b + 1) + b)

        if self.render_args.get("cell").obscured():
            output += AnsiImageRenderer.select().render(
                data, width=self.width, height=self.height
            )
        else:
            # Delete existing image in this space
            # output += f"\001\x1b_Ga=d,d=i,i={image_id}\x1b\\\002"
            # Fill image space
            output += "\n".join([" " * self.width] * self.height)
            # Save cursor position
            output += "\x1b[s"
            # Move back to start of image position
            output += f"\x1b[{self.height-1}A\x1b[{self.width}D"
            # Send the image
            params = {
                "a": "T",  # We are sending an image
                "q": 1,  # No backchat
                "f": "100",  # Sending a PNG image
                "t": "d",  # Transferring the image directly
                "i": image_id,
                # 'p': 1,
                "m": 1,  # Data will be chunked
                "c": self.width,
                "r": self.height,
                "z": -(2 ** 30) - 1,  # Put image under everything
            }
            # Remove newlines from base64 image data
            data = "".join(data.split("\n"))
            while data:
                chunk, data = data[:4096], data[4096:]
                params["m"] = 1 if data else 0
                param_str = ",".join(
                    [f"{key}={value}" for key, value in params.items()]
                )
                output += f"\001\x1b_G{param_str};{chunk}\x1b\\\002"
            # Restore cursor position
            output += "\x1b[u"
        return output


class AnsiImageRenderer(ImageRenderer):
    def process(self, data):
        data = base64.b64decode(data)
        # if hasattr(super(), 'process'):
        data = super().process(data)
        return data
        # return data


class img_ansi_timg(AnsiImageRenderer, ImageRenderer, SubprocessRenderMixin):
    def validate(self):
        # There is a python package called timg which provided an executable
        # That's not the one we are aftere here, so check if it is installed
        try:
            import timg  # noqa F401
        except ModuleNotFoundError:
            return super().validate()
        else:
            return False

    def load(self):
        self.cmd = [
            "timg",
            f"-g{self.width}x{self.width}",
            "--compress",
            "-b",
            get_app().bg_color,
            "-pq",
            "--threads=-1",
            "-",
        ]


class img_ansi_catimg(AnsiImageRenderer, ImageRenderer, SubprocessRenderMixin):
    def load(self):
        self.cmd = ["catimg", "-w", self.width * 2, "-"]


class img_ansi_icat(AnsiImageRenderer, ImageRenderer, SubprocessRenderMixin):
    use_tempfile = True

    def load(self):
        self.cmd = ["icat", "-w", self.width, "--mode", "24bit"]


class ansi_tiv(AnsiImageRenderer, ImageRenderer, SubprocessRenderMixin):
    use_tempfile = True

    def load(self):
        self.cmd = ["tiv", "-w", self.width, "-h", self.height]


class img_ansi_timg_py(AnsiImageRenderer, ImageRenderer, PythonRenderMixin):
    module = "timg"

    def process(self, data):
        import timg

        self.image.thumbnail((self.width * 2, self.height * 2))
        # Set transparent colour to terminal background
        if (
            bg_color := get_app().bg_color
            and self.image.mode in ("RGBA", "LA")
            or (self.image.mode == "P" and "transparency" in self.image.info)
        ):
            alpha = self.image.convert("RGBA").getchannel("A")
            bg = Image.new("RGBA", self.image.size, bg_color)
            bg.paste(self.image, mask=alpha)
            self.image = bg
        self.image = self.image.convert("P", palette=Image.ADAPTIVE, colors=16).convert(
            "RGB", palette=Image.ADAPTIVE, colors=16
        )
        data = timg.Ansi24HblockMethod(self.image).to_string()
        return data


class img_ansi_img2unicode_py(AnsiImageRenderer, ImageRenderer, PythonRenderMixin):
    module = "img2unicode"

    def process(self, data):
        from img2unicode import FastQuadDualOptimizer, Renderer

        output = io.StringIO()
        Renderer(
            FastQuadDualOptimizer(), max_w=self.width, max_h=self.height
        ).render_terminal(self.image, output)
        output.seek(0)
        return output.read()


class img_ansi_viu(AnsiImageRenderer, ImageRenderer, SubprocessRenderMixin):
    def load(self):
        self.cmd = ["viu", "-w", self.width, "-s", "-"]


class img_ansi_jp2a(AnsiImageRenderer, ImageRenderer, SubprocessRenderMixin):
    def load(self):
        self.cmd = ["jp2a", "--color", f"--height={self.height}", "-"]


class img_ansi_img2txt(AnsiImageRenderer, ImageRenderer, SubprocessRenderMixin):
    use_tempfile = True

    def load(self):
        self.cmd = ["img2txt", "-W", self.width, "-H", self.height]


class img_ansi_placeholder(AnsiImageRenderer, ImageRenderer):
    msg = "[Image]"

    def validate(self):
        # This is always an option
        return True

    def process(self, data):
        b = Border
        t = len(self.msg)
        w = max(self.width, t + 4)
        h = max(3, self.height)
        output = b.TOP_LEFT + (b.HORIZONTAL * (w - 2)) + b.TOP_RIGHT + "\n"
        for i in range((h - 3) // 2 + 1):
            output += b.VERTICAL + (" " * (w - 2)) + b.VERTICAL + "\n"
        output += b.VERTICAL + " "
        output += " " * ((self.width - 4 - t) // 2)
        output += self.msg
        output += " " * ((self.width - 4 - t) - (self.width - 4 - t) // 2)
        output += " " + b.VERTICAL + "\n"
        for i in range((h - 3) - ((h - 3) // 2)):
            output += b.VERTICAL + (" " * (w - 2)) + b.VERTICAL + "\n"
        output += b.BOTTOM_LEFT + (b.HORIZONTAL * (w - 2)) + b.BOTTOM_RIGHT
        return output


class SVGRenderer(DataRenderer):
    pass


class svg_librsvg(SVGRenderer, PythonRenderMixin):
    module = "cairosvg"

    def process(self, data):
        import cairosvg

        data = cairosvg.surface.PNGSurface.convert(data, write_to=None)
        data = base64.b64encode(data)
        return ImageRenderer.select().render(
            data, self.width, self.height, self.render_args
        )


class svg_imagemagik(SVGRenderer, SubprocessRenderMixin):
    to_string = False

    def load(self):
        self.cmd = [
            "convert",
            "-",
            "PNG:-",
        ]

    def process(self, data):
        data = super().process(data)
        data = base64.b64encode(data)
        return ImageRenderer.select().render(
            data, self.width, self.height, self.render_args
        )
